Combo.Socket.Message (combo v0.8.0)
View SourceDefines a message dispatched over transport to channels and vice-versa.
The message format requires the following keys:
:topic- The string topic or topic:subtopic pair namespace, for example "messages", "messages:123":event- The string event name, for example "combo_join":payload- The message payload:ref- The unique string ref:join_ref- The unique string ref when joining
Summary
Functions
Converts a map with string keys into a message struct.
Types
Functions
Converts a map with string keys into a message struct.
Raises Combo.Socket.InvalidMessageError if not valid.