Combo.Template.CEExEngine (combo v0.10.0)

The template engine that handles CEEx templates.

CEEx is a template language forked from Phoenix's HEEx, which strips out code related to LiveView, aiming solely at static template rendering.

CEEx stands for "Combo EEx". Maybe the name isn't very descriptive, but at least it clearly distinguishes itself from HEEx.

Features

  • syntax checking
  • assigns enhancement:
    • @ syntax
    • declarative assigns
  • component system
  • protection on XSS

Modules

The core feature is implemented by Combo.Template.CEExEngine.Compiler.

And, other additional features are implemented by following modules:

In practice, we rarely use these modules directly. Instead, we use Combo.HTML which is built on top of them.

Summary

Functions

Quick setup for using this engine.

Functions

__using__(opts \\ [])

(macro)

Quick setup for using this engine.