mix combo.gen.secret (combo v0.10.0)
View SourceGenerates a secret and prints it to the stdout.
$ mix combo.gen.secret [length]
The default value for length is 64.
The minimum value for length is 32.
Generates a secret and prints it to the stdout.
$ mix combo.gen.secret [length]
The default value for length is 64.
The minimum value for length is 32.